import random

from typing import List, Optional
import log

logger = log.setup_logger(__name__)


class ChatNode:
  id: str
  content: str
  role: str

  parent: Optional['ChatNode']
  children: List['ChatNode']

  visits: int
  value: float
  meta: dict

  def from_conversation(messages):
    root_message = messages[0]
    node = ChatNode(role=root_message['role'], content=root_message['content'])

    for message in messages[1:]:
      node = node.add_child(
        ChatNode(role=message['role'], content=message['content'])
      )

    return node

  def __init__(self, **kwargs):
    self.id = ''.join(
      random.choices('abcdefghijklmnopqrstuvwxyz0987654321', k=4)
    )
    self.content = kwargs.get('content', '')
    self.role = kwargs.get('role', '')

    self.parent = kwargs.get('parent', None)
    self.children = kwargs.get('children', [])

    self.visits = kwargs.get('visits', 0)
    self.value = kwargs.get('value', 0.0)

    self.meta = kwargs.get('meta', {})

  def add_parent(self, parent: 'ChatNode'):
    parent.children.append(self)
    self.parent = parent
    return self

  def add_child(self, child: 'ChatNode'):
    child.parent = self
    self.children.append(child)
    return child

  def best_child(self):
    if not self.children:
      return self
    return max(self.children, key=lambda c: c.value).best_child()

  def contains(self, substring):
    return substring.lower() in self.content.lower()

  def parents(self):
    parents = [self]

    while self.parent:
      self = self.parent
      parents.append(self)

    return parents[::-1]

  def message(self):
    return {
      "role": self.role,
      "content": self.content,
    }

  def ancestor(self):
    node = self
    while node.parent:
      node = node.parent
    return node

  def history(self):
    node = self
    messages = [node.message()]

    while node.parent:
      node = node.parent
      messages.append(node.message())

    return messages[::-1]

  def __str__(self):
    return f"{self.role}: {self.content}"



# Recursive Certainty Validation - RCN
# aka "Are you sure?

from config import RCN_STRAT, RCN_STRAT_PARAMS

import llm
import log
import chat as ch
import selection

logger = log.setup_logger(__name__)

ID_PREFIX = "rcn"


async def apply(chat: 'ch.Chat', llm: 'llm.LLM'):
  strat = RCN_STRAT.value
  strat_params = RCN_STRAT_PARAMS.value
  debug_info = {
    "strat": strat,
    "strat_params": strat_params,
  }

  logger.debug(f"{ID_PREFIX}: {debug_info}")
  nodes = selection.apply_strategy(chat, strategy=strat, params=strat_params)

  if (len(nodes) > 1):
    logger.warning(
      f"{ID_PREFIX}: Matched multiple nodes, only the first one will be processed."
    )

  if len(nodes) == 0:
    log.info(f"{ID_PREFIX}: No nodes matched, skipping.")
    return llm.stream_chat_completion(chat)

  node = nodes[0]
  question = node.content

  output = chat.Chat.from_conversation(
    [
      {
        "role":
          "system",
      "content":
          """
YOU HAVE LIMITATIONS AS AN LLM. DO NOT OVERCOMPLICATE THINGS. YOU MAKE MISTAKES ALL THE TIME, SO BE CAREFUL IN YOUR REASONING.
WHEN SOLVING PROBLEMS - DECOMPOSE THEM INTO SMALLER PARTS. SOLVE PARTS ONE BY ONE SEQUENTIALLY.
DECLARE THE INITIAL STATE, MODIFY IT ONE STEP AT A TIME. CHECK THE RESULT AFTER EACH MODIFICATION.
DO NOT SAY YOU DOUBLE-CHECKED AND TRIPLE-CHECKED WITHOUT ACTUALLY DOING SO.
""".strip()
      }, {
        "role":
          "user",
        "content":
          f"""
Take this question:
{question}

Describe the meaning of every word in relation to the question. Paraphrase the question two times. Then provide a solution.
""".strip()
      }
    ]
  )
  output.llm = llm

  await output.advance()
  output.user("Are you sure?")
  await output.advance()
  output.user("Is this yout final answer?")
  await output.advance()
  output.user(
    "Now prepare your final answer. Write it as a response to this message. Do not write anything else."
  )

  await llm.stream_final_completion(chat=output)



"""
title: mcts
author: av
author_url: https://github.com/av
description: mcts - Monte Carlo Tree Search
version: 0.0.5
"""

import logging
import random
import math
import asyncio
import json
import re

from typing import (
  List,
  Optional,
  AsyncGenerator,
  Callable,
  Awaitable,
  Generator,
  Iterator,
)
from open_webui.constants import TASKS
from open_webui.apps.ollama import main as ollama

# ==============================================================================

name = "mcts"
default_max_children = 2
default_exploration_weight = 1.414
default_max_iterations = 2
default_max_simulations = 2
default_thoughts = 2

# ==============================================================================

thoughts_prompt = """
<instruction>
Give a suggestion on how this answer can be improved.
WRITE ONLY AN IMPROVEMENT SUGGESTION AND NOTHING ELSE.
YOUR REPLY SHOULD BE A SINGLE SENTENCE.
</instruction>

<question>
{question}
</question>

<draft>
{answer}
</draft>
""".strip()

eval_answer_prompt = """
Given the following text:
"{answer}"

How well does it answers this question:
"{question}"

Rate the answer from 1 to 10, where 1 is completely wrong or irrelevant and 10 is a perfect answer.
Reply with a single number between 1 and 10 only. Do not write anything else, it will be discarded.
THINK CAREFULLY AND USE BEST PRACTICES.
""".strip()

analyze_prompt = """
Iteration Analysis:

Original question: {question}
Best answer found: {best_answer}
Best score achieved: {best_score}

Analyze this iteration of the thought process. Consider the following:
1. What aspects of the best answer made it successful?
2. What patterns or approaches led to higher-scoring thoughts?
3. Were there any common pitfalls or irrelevant tangents in lower-scoring thoughts?
4. How can the thought generation process be improved for the next iteration?

Provide a concise analysis and suggest one specific improvement strategy for the next iteration.
""".strip()

update_prompt = """
<instruction>
Your task is to read the question and the answer below, then analyse the given critique.
When you are done - think about how the answer can be improved based on the critique.
WRITE A REVISED ANSWER THAT ADDRESSES THE CRITIQUE. DO NOT WRITE ANYTHING ELSE.
</instruction>
<question>
{question}
</question>
<draft>
{answer}
</draft>
<critique>
{improvements}
</critique>
""".strip()

initial_prompt = """
<instruction>
Answer the question below. Do not pay attention to, unexpected casing, punctuation or accent marks.
</instruction>

<question>
{question}
</question>
"""

# ==============================================================================


def setup_logger():
  logger = logging.getLogger(__name__)
  if not logger.handlers:
    logger.setLevel(logging.DEBUG)
    handler = logging.StreamHandler()
    handler.set_name(name)
    formatter = logging.Formatter(
      "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.propagate = False
  return logger


logger = setup_logger()

# ==============================================================================

mods = [
  "capitalize",
  "diacritic",
  "leetspeak",
  "remove_vowel",
]


def modify_text(text, percentage):
  if not text:
    return "", {}    # Return empty string and empty mapping if input is empty

  if not 0 <= percentage <= 100:
    raise ValueError("Percentage must be between 0 and 100")

  words = text.split()
  chars = list(text)
  num_chars_to_modify = max(1, int(len(chars) * (percentage / 100)))
  indices_to_modify = random.sample(range(len(chars)), num_chars_to_modify)
  word_mapping = {}

  for idx in indices_to_modify:
    modification = random.choice(mods)

    # Find the word that contains the current character
    current_length = 0
    for word_idx, word in enumerate(words):
      if current_length <= idx < current_length + len(word):
        original_word = word
        word_start_idx = current_length
        break
      current_length += len(word) + 1    # +1 for the space
    else:
      # If we're here, we're likely dealing with a space or the last character
      continue

    if modification == "capitalize":
      chars[idx] = chars[idx].swapcase()
    elif modification == "diacritic":
      if chars[idx].isalpha():
        diacritics = ["̀", "́", "̂", "̃", "̈", "̄", "̆", "̇", "̊", "̋"]
        chars[idx] = chars[idx] + random.choice(diacritics)
    elif modification == "leetspeak":
      leetspeak_map = {
        "a": "4",
        "e": "3",
        "i": "1",
        "o": "0",
        "s": "5",
        "t": "7",
        "b": "8",
        "g": "9",
        "l": "1",
      }
      chars[idx] = leetspeak_map.get(chars[idx].lower(), chars[idx])
    elif modification == "remove_vowel":
      if chars[idx].lower() in "aeiou":
        chars[idx] = ""

    modified_word = "".join(
      chars[word_start_idx:word_start_idx + len(original_word)]
    )

    if modified_word != original_word:
      # Clean up both the modified word and the original word
      cleaned_modified_word = modified_word.rstrip(".,!?")
      cleaned_original_word = original_word.rstrip(".,!?")
      word_mapping[cleaned_modified_word] = cleaned_original_word

  modified_text = "".join(chars)
  return modified_text, word_mapping


def replace_with_mapping(text, mapping):
  for key, value in mapping.items():
    text = text.replace(key, value)
  return text


# ==============================================================================


def escape_mermaid(text):
  return text.replace('"', "&quot;").replace("(", "&#40;").replace(")", "&#41;")


class Node:
  id: str
  content: str
  parent: Optional["Node"]
  max_children: int
  children: List["Node"]
  visits: int
  value: float

  def __init__(self, **kwargs):
    self.id = "".join(random.choices("abcdefghijklmnopqrstuvwxyz", k=4))
    self.content = kwargs.get("content")
    self.parent = kwargs.get("parent")
    self.exploration_weight = kwargs.get(
      "exploration_weight", default_exploration_weight
    )
    self.max_children = kwargs.get("max_children", default_max_children)
    self.children = []
    self.visits = 0
    self.value = 0

  def add_child(self, child: "Node"):
    child.parent = self
    self.children.append(child)
    return child

  def fully_expanded(self):
    return len(self.children) >= self.max_children

  def uct_value(self):
    epsilon = 1e-6

    return self.value / (self.visits +
                         epsilon) + self.exploration_weight * math.sqrt(
                           math.log(self.parent.visits) /
                           (self.visits + epsilon)
                         )

  def mermaid(self, offset=0, selected=None):
    padding = " " * offset
    msg = f"{padding}{self.id}({self.id}:{self.visits} - {escape_mermaid(self.content[:25])})\n"

    if selected == self.id:
      msg += f"{padding}style {self.id} stroke:#0ff\n"

    for child in self.children:
      msg += child.mermaid(offset + 4, selected)
      msg += f"{padding}{self.id} --> {child.id}\n"

    return msg

  def best_child(self):
    if not self.children:
      return self

    return max(self.children, key=lambda child: child.visits).best_child()


class MCTS:
  question: str
  root: Node
  llm: "Pipe"
  selected: Optional[Node]
  exploration_weight: float

  def __init__(self, **kwargs):
    self.question = kwargs.get("question")
    self.root = kwargs.get("root")
    self.llm = kwargs.get("llm")
    self.selected = None
    self.exploration_weight = kwargs.get(
      "exploration_weight", default_exploration_weight
    )

  async def select(self):
    logger.debug("Selecting node...")
    node = self.root
    while node.children:
      node = self.uct_select(node)
    return node

  async def expand(self, node):
    logger.debug(f"Expanding node {node.id}...")
    await self.llm.progress(f"Thinking about {node.id}...")

    for _ in range(random.randint(default_thoughts, default_thoughts + 1)):
      await self.llm.emit_replace(self.mermaid(node))
      await self.llm.emit_message(f"Thought: ")
      thought = await self.llm.generate_thought(node.content)
      await self.llm.emit_message(f"\n\n---\n\nSolution:\n")

      new_content = await self.llm.update_approach(node.content, thought)
      child = Node(content=new_content, parent=node)
      node.add_child(child)

    return random.choice(node.children)

  async def simulate(self, node):
    logger.debug(f"Simulating node {node.id}...")
    await self.llm.progress(f"Thinking about {node.id}...")
    await self.llm.emit_replace(self.mermaid())

    return await self.llm.evaluate_answer(node.content)

  def backpropagate(self, node, score):
    logger.debug(f"Backpropagating from {node.id}...")
    while node:
      node.visits += 1
      node.value += score
      node = node.parent

  def uct_select(self, node):
    logger.debug(f"Selecting uct {node.id}...")
    return max(node.children, key=lambda child: child.uct_value())

  def best_child(self):
    return self.root.best_child()

  async def search(self, num_simulations):
    logger.debug("Starting search...")

    for _ in range(num_simulations):
      leaf = await self.select()
      self.selected = leaf
      if not leaf.fully_expanded():
        leaf = await self.expand(leaf)
      score = await self.simulate(leaf)
      self.backpropagate(leaf, score)

    return self.selected

  def mermaid(self, selected=None):
    return f"""
```mermaid
graph LR
{self.root.mermaid(0, selected.id if selected else self.selected.id)}
```
"""


# ==============================================================================

EventEmitter = Callable[[dict], Awaitable[None]]


class Pipe:
  __current_event_emitter__: EventEmitter
  __current_node__: Node
  __question__: str
  __model__: str

  def __init__(self):
    self.type = "manifold"

  def pipes(self) -> list[dict[str, str]]:
    ollama.get_all_models()
    models = ollama.app.state.MODELS

    out = [
      {
        "id": f"{name}-{key}",
        "name": f"{name} {models[key]['name']}"
      } for key in models
    ]
    logger.debug(f"Available models: {out}")

    return out

  def resolve_model(self, body: dict) -> str:
    model_id = body.get("model")
    without_pipe = ".".join(model_id.split(".")[1:])
    return without_pipe.replace(f"{name}-", "")

  def resolve_question(self, body: dict) -> str:
    return body.get("messages")[-1].get("content").strip()

  async def pipe(
    self,
    body: dict,
    __user__: dict,
    __event_emitter__=None,
    __task__=None,
    __model__=None,
  ) -> str | Generator | Iterator:
    model = self.resolve_model(body)
    base_question = self.resolve_question(body)

    if __task__ == TASKS.TITLE_GENERATION:
      content = await self.get_completion(model, body.get("messages"))
      return f"{name}: {content}"

    logger.debug(f"Pipe {name} received: {body}")
    question, mapping = modify_text(base_question, 0)
    logger.debug(f"Question: {question}")

    # TODO: concurrency
    self.__model__ = model
    self.__question__ = base_question
    self.__current_event_emitter__ = __event_emitter__

    best_answer = None
    best_score = -float("inf")

    await self.progress("Preparing initial thoughts...")
    initial_reply = await self.stream_prompt_completion(
      initial_prompt, question=question
    )

    root = Node(content=initial_reply)
    mcts = MCTS(root=root, llm=self)

    logger.debug("Starting MCTS...")
    for i in range(default_max_iterations):
      logger.debug(f"Iteration {i + 1}/{default_max_iterations}...")

      await mcts.search(default_max_simulations)
      logger.debug(mcts.mermaid())

      best_child = mcts.best_child()
      score = await self.evaluate_answer(best_child.content)

      if score > best_score:
        best_score = score
        best_answer = best_child.content

    await self.emit_replace(mcts.mermaid(best_child))
    await self.emit_message(f"{best_answer}")
    await asyncio.sleep(0.2)
    await self.done()

    return ""

  async def progress(
    self,
    message: str,
  ):
    logger.debug(f"Progress: {message}")
    await self.emit_status("info", message, False)

  async def done(self,):
    await self.emit_status("info", "Fin.", True)

  async def emit_message(self, message: str):
    await self.__current_event_emitter__(
      {
        "type": "message",
        "data": {
          "content": message
        }
      }
    )

  async def emit_replace(self, message: str):
    await self.__current_event_emitter__(
      {
        "type": "replace",
        "data": {
          "content": message
        }
      }
    )

  async def emit_status(self, level: str, message: str, done: bool):
    await self.__current_event_emitter__(
      {
        "type": "status",
        "data":
          {
            "status": "complete" if done else "in_progress",
            "level": level,
            "description": message,
            "done": done,
          },
      }
    )

  async def get_streaming_completion(
    self,
    model: str,
    messages,
  ) -> AsyncGenerator[str, None]:
    response = await ollama.generate_openai_chat_completion(
      {
        "model": model,
        "messages": messages,
        "stream": True
      }
    )

    async for chunk in response.body_iterator:
      for part in self.get_chunk_content(chunk):
        yield part

  async def get_message_completion(self, model: str, content):
    async for chunk in self.get_streaming_completion(
      model, [{
        "role": "user",
        "content": content
      }]
    ):
      yield chunk

  async def get_completion(self, model: str, messages):
    response = await ollama.generate_openai_chat_completion(
      {
        "model": model,
        "messages": messages,
        "stream": False
      }
    )

    return self.get_response_content(response)

  async def stream_prompt_completion(self, prompt, **format_args):
    complete = ""
    async for chunk in self.get_message_completion(
      self.__model__,
      prompt.format(**format_args),
    ):
      complete += chunk
      await self.emit_message(chunk)
    return complete

  async def generate_thought(self, answer):
    return await self.stream_prompt_completion(
      thoughts_prompt, answer=answer, question=self.__question__
    )

  async def analyze_iteration(self, best_answer, best_score):
    return await self.stream_prompt_completion(
      analyze_prompt,
      question=self.__question__,
      best_answer=best_answer,
      best_score=best_score
    )

  async def update_approach(self, answer, improvements):
    return await self.stream_prompt_completion(
      update_prompt,
      question=self.__question__,
      answer=answer,
      improvements=improvements
    )

  async def evaluate_answer(self, answer):
    result = await self.stream_prompt_completion(
      eval_answer_prompt,
      answer=answer,
      question=self.__question__,
    )
    try:
      score = re.search(r"\d+", result).group()
      return int(score)
    except AttributeError:
      logger.error(f"AnswerEval: unable to parse \"{result[:100]}\"")
      return 0

  def get_response_content(self, response):
    try:
      return response["choices"][0]["message"]["content"]
    except (KeyError, IndexError):
      logger.error(
        f"ResponseError: unable to extract content from \"{response[:100]}\""
      )
      return ""

  def get_chunk_content(self, chunk):
    chunk_str = chunk.decode("utf-8")
    if chunk_str.startswith("data: "):
      chunk_str = chunk_str[6:]

    chunk_str = chunk_str.strip()

    if chunk_str == "[DONE]" or not chunk_str:
      return

    try:
      chunk_data = json.loads(chunk_str)
      if "choices" in chunk_data and len(chunk_data["choices"]) > 0:
        delta = chunk_data["choices"][0].get("delta", {})
        if "content" in delta:
          yield delta["content"]
    except json.JSONDecodeError:
      logger.error(f"ChunkDecodeError: unable to parse \"{chunk_str[:100]}\"")



import llm
import log
import chat as ch

ID_PREFIX = 'example'
logger = log.setup_logger(ID_PREFIX)


async def apply(chat: 'ch.Chat', llm: 'llm.LLM'):
  """
  1. Working with a chat and chat nodes instances
  This is where you can create some content programmatically,
  that will later can be used for retrieving completions from
  the downstream model.
  """
  logger.debug(f"Example chat: {chat}")

  # Add new messages to the chat (no completions at this stage)
  chat.user("Hello!")
  chat.assistant("Hi! Would you like to learn more about Harbor Boost?")
  chat.add_message(
    role="harbor",
    content="Harbor Boost is an optimising LLM proxy with lots of cool features"
  )

  logger.debug(
    f"Chat history is a plain array of messages, from the tail: {chat.history()}"
  )
  logger.debug(
    f"Chat plain is a list of chat nodes, from the tail: {chat.plain()}"
  )

  # Tail is where the chat currently ends
  # In this instance, that's a message from "harbor"
  # role above
  tail = chat.tail

  logger.debug(
    f'Get all parents leading to a specific chat node: {tail.parents()}'
  )
  logger.debug(f'Get one immediate parent: {tail.parent}')

  # We can modify the chat from the tail node directly
  new_tail = tail.add_child(
    ch.ChatNode(role="harbor", content="Chat nodes are everywhere!")
  )

  # However, such modifications are not reflected in the parent
  # chat instance:
  logger.debug(chat.tail == tail)    # True
  logger.debug(chat.tail == new_tail)    # False

  # You can set a new tail for the chat, however
  chat.tail = new_tail
  # However, it's much easier to just work from the chat itself
  chat.user('Alright, I think that is mostly it for now. Thanks!')

  # You can create new chat instances as needed
  life_chat = ch.Chat.from_conversation(
    [
      {
        "role": "user",
        "content": "What is the meaning of life? Answer with a tongue twister."
      }
    ]
  )
  """
  3.1 Programmatic messages and statuses
  programmatic "public" messages that are streamed
  back to the client as they are emitted here
  (no way to "undo" or rewrite them)
  """
  # You can tweak how status messages are delivered
  # via the BOOST_STATUS_STYLE config option.
  await llm.emit_status('Status and message examples')
  await llm.emit_message("We can emit text at any time. ")
  await llm.emit_message(
    "\n_Note that you are responsible for correct formatting._\n"
  )
  """
  3.2. Internal LLM completions
  "llm" is a representation of the downstream model
  that is being boosted. It comes with a few helper
  methods that tie up the module workflow together and
  is pre-configured to hit the downstream API with expected parameters.

  The completions below are "internal", they are not streamed
  back to the client by default. Read further for "streamed" or
  "public" completions.
  """
  await llm.emit_status('Collecting internal completions...')
  word = "Roses"
  results = [
    # You can retrieve completion for some plain text
    await llm.chat_completion(prompt="Hi!", resolve=True),
    # You can include key/value pairs to be formatted in the prompt
    await llm.chat_completion(
      prompt="Tell me about {word} in ONE SHORT SENTENCE.",
      word=word,
      resolve=True,
    ),
    # You can also provide a list of messages
    # in the OpenAI-compatible format
    await llm.chat_completion(
      messages=[
        {
          "role": "user",
          "content": "Tell me about roses"
        }, {
          "role": "assistant",
          "content": "Sure, I can reply in three words! Here they are:"
        }
      ],
      resolve=True
    ),
    # You can also provide a chat instance,
    # Note that "resolve" is not set - the result
    # will be in raw API format
    f"\n```json\n{await llm.chat_completion(chat=life_chat)}\n```\n"
  ]
  # Results will now appear in the user's message
  await llm.emit_status('Displaying collected results')
  for i, result in enumerate(results):
    await llm.emit_message(f"\nResult {i}: {result}\n")
  """
  3.3. Public/Streamed LLM completions
  You can decide to stream responses from the downstream LLM
  as they are being generated, for example when there's a long
  chunk that needs to be retained in the global response.
  """
  await llm.emit_status('Response streaming examples')

  # Same signatures as chat_completion
  streamed_results = [
    # You can retrieve completion for some plain text
    await llm.stream_chat_completion(prompt="Hi!"),
    # You can include key/value pairs to be formatted in the prompt
    await llm.stream_chat_completion(
      prompt="Tell me about {word} in ONE SHORT SENTENCE.",
      word=word,
    ),
    # You can also provide a list of messages
    # in the OpenAI-compatible format
    await llm.stream_chat_completion(
      messages=[
        {
          "role": "user",
          "content": "Tell me about roses"
        }, {
          "role": "assistant",
          "content": "Sure, I can reply in three words! Here they are:"
        }
      ],
    ),
    # You can also provide a chat instance
    await llm.stream_chat_completion(chat=life_chat)
  ]
  # Streamed results are still buffered and available
  # for you to use (plain text).
  logger.debug(f"Streamed results: {streamed_results}")

  # Note that it's on you to apply formatting that will make
  # sense in the context of the global message stream.
  await llm.emit_message("\nThose are all results so far.\n")
  """
  4. Final completion
  Note that none of the above will actually reach the Client
  if the BOOST_INTERMEDIATE_OUTPUT is set to "false".
  The "final" completion below, however, will *always* be streamed back.
  It accepts all the same inputs as "chat_completion" and "stream_chat_completion" above.
  You don't have to call it, but the output will be completely empty if the
  "final" completion is not called and intermediate outputs are disabled.

  Think of this as a way to wrap up the module execution and
  present the user with the final result.
  """
  await llm.emit_status('Final completion')
  await llm.stream_final_completion(prompt="Wish me a good luck!",)



import chat as ch
import llm

ID_PREFIX = 'clarity'

should_clarify_prompt = """
<instruction>
Is this question requires any clarification or is ready to be answered?
Reply only with "clarify" or "ready" and nothing else. Everything else will be ignored.
</instruction>

<question>
{question}
</question>
  """.strip()


async def apply(chat: 'ch.Chat', llm: 'llm.LLM'):
  iterations = 0
  max_iterations = 15

  side_chat = ch.Chat.from_conversation([chat.tail.message()])
  side_chat.llm = llm

  while iterations < max_iterations:
    iterations += 1
    side_chat.user(
      """
Are there any sources of ambiguity in my request?
Answer with "yes" or "no" and nothing else. Everything else will be ignored.
    """.strip()
    )
    await side_chat.advance()
    await llm.emit_status(f'Clarification: {side_chat.tail.content}')

    if side_chat.tail.contains('no'):
      break

    side_chat.user("""
Clarify the ambiguity you mentioned.
    """.strip())
    await side_chat.emit_advance()

    if iterations >= max_iterations:
      break

  side_chat.user('Now, please provide a clear answer to the question.')
  await side_chat.emit_advance()

  await llm.emit_status('Final')

  side_chat.user(
    """
Think trough the response you just gave. Is there anything wrong? If so, please correct it.
Otherwise, write down your final answer to my request.
    """.strip()
  )
  await llm.stream_final_completion(chat=chat)


import chat as ch
import llm

ID_PREFIX = 'ambi'

ambi_prompt = """
<instruction>
Find the sources of ambiguities in the given question and describe them.
</instruction>

<question>
{question}
</question>
  """.strip()

detail_prompt = """
<instruction>
Find the conditions that significantly affect the interpretation of the question and describe them.
</instruction>

<question>
{question}
</question>
""".strip()

definition_prompt = """
<instruction>
Define the terms in the question and provide a detailed explanation for each.
</instruction>

<question>
{question}
</question>
""".strip()

discrepancies_prompt = """
<instruction>
Find the discrepancies in the question and describe them.
</instruction>

<question>
{question}
</question>
""".strip()

final_prompt = """
<instruction>
Provide a clear and definitive answer to the question.
</instruction>

<question>
{question}
</question>

<information>
### Ambiguities
{ambiguities}

### Details
{details}

### Definitions
{definitions}

### Discrepancies
{discrepancies}
</information>
""".strip()


async def apply(chat: 'ch.Chat', llm: 'llm.LLM'):
  await llm.emit_status('Ambiguiity')
  ambiguities = await llm.stream_chat_completion(
    prompt=ambi_prompt, question=chat.tail.content
  )

  await llm.emit_status('Details')
  details = await llm.stream_chat_completion(
    prompt=detail_prompt, question=chat.tail.content
  )

  await llm.emit_status('Definitions')
  definitions = await llm.stream_chat_completion(
    prompt=definition_prompt, question=chat.tail.content
  )

  await llm.emit_status('Discrepancies')
  discrepancies = await llm.stream_chat_completion(
    prompt=discrepancies_prompt, question=chat.tail.content
  )

  await llm.emit_status('Final')
  await llm.stream_final_completion(
    prompt=final_prompt,
    question=chat.tail.content,
    ambiguities=ambiguities,
    details=details,
    definitions=definitions,
    discrepancies=discrepancies
  )



import random

import chat as ch
import llm
import log

ID_PREFIX = 'pad'

THINKING = [
  'Actively thinking...',
  'Considering possible implications...',
  'Simplifying...',
  'Ensuring correctness...',
  'Clarifying...',
  'Rephrasing a thought...',
  'Reconsidering...',
  'Evaluating assumption...',
  'Analyzing constraints...',
  'Reflecting on a solution...',
  'Reviewing...',
  'Contemplating...',
  'Pondering...',
  'Speculating...',
  'Deliberating...',
  'Focusing on the outcome...',
  'Imagining alternatives...',
  'Envisioning a simpler path...',
  'Creating a more concise outline...',
  'Constructing a more elaborate plan...',
  'Designing a more efficient solution...',
  'Inventing a more effective strategy...',
  'Devising a more practical approach...',
  'Formulating a more sophisticated method...',
  'Developing a more advanced technique...',
]

THINKING_2 = [
  'Thinking about task at hand',
  'Applying critical thinking',
  'Choosing more practical options',
  'Ensuring pragmatic solutions',
  'Simplifying',
  'Ensuring correctness',
  'Ensuring practicality'
  'Removing sources of ambiguity',
  'Making sure it makes sense',
  'Ensuring clarity',
  'Ensuring simplicity',
  'Making sure it\'s easy to understand',
  'Verifying the logic',
  'Checking for errors',
  'Making sure I did not miss anything',
  'Avoiding obvious mistakes',
  'Fixing an error',
  'Correcting a mistake',
  'Breaking down the problem',
  'Ensuring the solution is feasible',
  'Clarifying assumptions',
]

WORDS = [
  'apple', 'banana', 'cherry', 'date', 'elderberry', 'fig', 'grape', 'honeydew',
  'kiwi', 'lemon', 'mango', 'nectarine', 'orange', 'pear', 'quince',
  'raspberry', 'strawberry', 'tangerine', 'ugli', 'vanilla', 'watermelon',
  'ximenia', 'yuzu', 'zucchini'
]


def get_size(**kwargs):
  return int(kwargs.get('pad_size', 256))


def pad_thinking(**kwargs):
  size = get_size(**kwargs)
  join = kwargs.get('join', '\n')
  return join.join([random.choice(THINKING_2) for _ in range(size)])


def pad_newline(**kwargs):
  size = get_size(**kwargs)
  return '\n' * size


def pad_space(**kwargs):
  size = get_size(**kwargs)
  return ' ' * size


def pad_random_nl(**kwargs):
  size = get_size(**kwargs)
  return ''.join([random.choice([' ', '\n']) for _ in range(size)])


def random_alphabet():
  return random.choice('abcdefghijklmnopqrstuvwxyz')


def pad_random_alphabet(**kwargs):
  size = get_size(**kwargs)
  join = kwargs.get('join', '')
  pad = join.join([random_alphabet() for _ in range(size)])
  return pad


def pad_random_words(**kwargs):
  size = get_size(**kwargs)
  pad = ' '.join([random.choice(WORDS) for _ in range(size)])
  return pad


def pad_random_numbers(**kwargs):
  size = get_size(**kwargs)
  pad = ' '.join([str(random.randint(0, 9)) for _ in range(size)])
  return pad


def pad_thinking_loop(**kwargs):
  size = get_size(**kwargs)
  join = kwargs.get('join', '\n')
  return join.join([THINKING_2[i % len(THINKING_2)] for i in range(size)])

def pad_thinking_steps(**kwargs):
  size = get_size(**kwargs)
  join = kwargs.get('join', '\n')

  return join.join([f'Step {i}: {THINKING_2[i % len(THINKING_2)]}...' for i in range(size)])

def pad_random_thinking_steps(**kwargs):
  size = get_size(**kwargs)
  join = kwargs.get('join', '\n')

  return join.join([f'Step {i}: {random.choice(THINKING_2)}...' for i in range(size)])


PAD_TYPES = {
  'thinking': pad_thinking,
  'newline': pad_newline,
  'space': pad_space,
  'random_nl': pad_random_nl,
  'random_alphabet': pad_random_alphabet,
  'random_words': pad_random_words,
  'random_numbers': pad_random_numbers,
  'thinking_loop': pad_thinking_loop,
  'thinking_steps': pad_thinking_steps,
  'random_thinking_steps': pad_random_thinking_steps,
}

PAD_STYLES = {
  'plain': lambda x: x,
  'block': lambda x: f"""
```entropy
{x}
```
""".strip(),
  'block_thoughts': lambda x: f"""
```thoughts
Starting internal thought process...
{x}
Ok, I am ready for the final answer now.
```
""".strip(),
  'quote': lambda x: f"""
> {x}
""".strip(),
}


def make_pad(**kwargs):
  pad_type = kwargs.get('pad_type', 'random_thinking_steps')
  pad_style = kwargs.get('pad_style', 'block_thoughts')

  return PAD_STYLES[pad_style](PAD_TYPES[pad_type](**kwargs))


logger = log.setup_logger(__name__)


async def apply(chat: 'ch.Chat', llm: 'llm.LLM'):
  llm.boost_params['pad_size'] = '128'

  pad = make_pad(**llm.boost_params)

  chat.user(
    f"""
Before addresing my request, I need you to take your time and think for a while.
It's very important for you to utilise this time to concentrate on the task at hand.
""".strip()
  )

  await chat.emit_status('Thinking...')

  chat.assistant(
    f"""
Thank you for letting me think for a bit! I will use this time to concentrate on the task at hand.
{pad}
  """.strip()
  )
  await llm.emit_message(chat.tail.content)

  chat.user(
    f"""
Ok, I think we're ready now. Please answer my previous request.
  """.strip()
  )

  await chat.emit_status('Final')
  await llm.stream_final_completion()



from typing import Optional

from chat_node import ChatNode
import llm
import log

logger = log.setup_logger(__name__)


class Chat:
  tail: ChatNode
  llm: Optional['llm.LLM']

  def from_conversation(messages):
    tail = ChatNode.from_conversation(messages)
    return Chat(tail=tail)

  def from_tail(chat):
    new_tail = ChatNode(role=chat.tail.role, content=chat.tail.content)
    return Chat(tail=new_tail)

  def __init__(self, **kwargs):
    self.tail = kwargs.get('tail')
    self.llm = kwargs.get('llm')
    self.chat_node_type = ChatNode

    self.Chat = Chat
    self.ChatNode = self.chat_node_type

  def has_substring(self, substring):
    return any(substring in msg.content for msg in self.plain())

  def add_message(self, role, content):
    logger.debug(f"Chat message: {role}: {content[:50]}")

    self.tail = self.tail.add_child(
      self.__create_node(role=role, content=content)
    )
    return self.tail

  def user(self, content):
    return self.add_message('user', content)

  def assistant(self, content):
    return self.add_message('assistant', content)

  def system(self, content):
    return self.add_message('system', content)

  def plain(self):
    return self.tail.parents()

  def history(self):
    return self.tail.history()

  def text(self):
    # __str__ already does exactly this
    return f"{self}"

  def __create_node(self, **kwargs):
    NodeType = self.chat_node_type
    return NodeType(**kwargs)

  async def advance(self):
    """
    Advance the chat completion

    Will not be streamed back to the client
    """

    if not self.llm:
      raise ValueError("Chat: unable to advance without an LLM")

    response = await self.llm.chat_completion(chat=self)
    self.assistant(self.llm.get_response_content(response))

  async def emit_advance(self):
    """
    Emit the next step in the chat completion

    Will be streamed back to the client
    """

    if not self.llm:
      raise ValueError("Chat: unable to advance without an LLM")

    response = await self.llm.stream_chat_completion(chat=self)
    self.assistant(response)

  async def emit_status(self, status):
    """
    Emit a status message

    Will be streamed back to the client
    """

    if not self.llm:
      raise ValueError("Chat: unable to emit status without an LLM")

    await self.llm.emit_status(status)

  def __str__(self):
    return '\n'.join([str(msg) for msg in self.plain()])



ID_PREFIX = "unstable"

import chat as ch
import llm
import random

extreme_traits = [
  "Eccentric", "Obsessive", "Impulsive", "Paranoid", "Narcissistic",
  "Perfectionist", "Overly Sensitive", "Extremely Independent", "Manipulative",
  "Aloof"
]

temperaments = ["Choleric", "Melancholic", "Phlegmatic", "Sanguine"]

reply_styles = [
  "Blunt", "Sarcastic", "Overly Polite", "Evading", "Confrontational"
]


# Function to generate a random personality description
def random_personality():
  selected_traits = random.sample(extreme_traits, 3)
  selected_temperament = random.choice(temperaments)
  selected_reply_style = random.choice(reply_styles)

  description = (
    f"You are {', '.join(selected_traits)}. "
    f"You are known for yout {selected_temperament} temperament."
    f"You tend to write your replies in a {selected_reply_style} manner."
    f"Ensure that you reply to the User accordingly."
  )

  return description


async def apply(chat: 'ch.Chat', llm: 'llm.LLM'):
  personality = random_personality()
  chat.tail.ancestor().add_parent(
    ch.ChatNode(role="system", content=personality)
  )

  await llm.stream_final_completion()



import random

import chat as ch
import llm
import log

ID_PREFIX = 'fml' # "formulaic", not what you think

async def apply(chat: 'ch.Chat', llm: 'llm.LLM'):
  chat.user(
    f"""
Rewrite my request in the formulaic logic language. Do not solve it yet.
    """.strip()
  )
  await chat.emit_status('Formulaic')
  await chat.emit_advance()

  chat.user(
    f"""
Solve my original request in the formulaic logic language.
""".strip()
  )
  await chat.emit_status('Solution')
  await chat.emit_advance()

  chat.user(
    f"""
Rewrite it in the natural language.
""".strip()
  )

  await chat.emit_status('Final')
  await llm.stream_final_completion()




import os
import subprocess
import sys

def main():
    current_dir = os.path.dirname(os.path.abspath(__file__))
    parent_dir = os.path.dirname(current_dir)
    harbor_sh_path = os.path.join(parent_dir, "harbor.sh")

    result = subprocess.run(
        ["bash", harbor_sh_path] + sys.argv[1:],
        shell=False,
        text=True,
        check=False,
        stdout=sys.stdout,
        stderr=sys.stderr,
    )
    sys.exit(result.returncode)

if __name__ == "__main__":
    main()


import random

import chat as ch
import llm

ID_PREFIX = 'wswp'

def swap_words(**args) -> str:
  text = args['text']
  probability = args['probability']

  words = text.split(' ')

  for i in range(len(words) - 1):
    if random.random() < probability:
      # Swap current word with the next one
      words[i], words[i + 1] = words[i + 1], words[i]

  return ' '.join(words)


async def apply(chat: 'ch.Chat', llm: 'llm.LLM'):
  chat.tail.content = swap_words(text=chat.tail.content, probability=0.15)
  await llm.emit_status(f'wswp: {chat.tail.content}')
  await llm.stream_final_completion()



import random
import re

from chat import Chat

def percentage(chat: Chat, **kwargs):
  percentage = kwargs.get("percentage", 50)
  nodes = chat.plain()
  num_nodes = max(1, int(len(nodes) * (percentage / 100)))

  return nodes[:num_nodes]


def match(chat: Chat, **kwargs):
  substring = kwargs.get("substring", "")
  role = kwargs.get("role", "")
  index = kwargs.get("index", None)

  nodes = chat.plain()

  if role:
    nodes = [node for node in nodes if node.role == role]

  if substring:
    nodes = [node for node in nodes if substring in node.content]

  if index is not None:
    nodes = [nodes[index]]

  return nodes


def user(chat: Chat):
  return match(chat, role="user")


def all(chat: Chat):
  return chat.plain()


def first(chat: Chat):
  return match(chat, index=0)


def last(chat: Chat):
  return match(chat, index=-1)


def any(chat: Chat):
  return [random.choice(chat.plain())]


selection_strategies = {
  "all": all,
  "first": first,
  "last": last,
  "any": any,
  "percentage": percentage,
  "match": match,
  "user": user,
}


def apply_strategy(chat: Chat, strategy: str, params: dict):
  return selection_strategies[strategy](chat, **params)

def match_regex(value, regex):
  return bool(re.match(regex, value))

def match_substring(value, substring):
  return substring in value

def match_exact(value, target):
  return value == target

def matches_filter(obj: dict, filter: dict):
  for key in filter.keys():
    value = filter[key]
    field, operation = key.split('.') if '.' in key else (key, 'exact')

    if field not in obj:
      return False

    if operation == 'regex':
      return match_regex(str(obj[field]), value)
    elif operation == 'contains':
      return match_substring(str(obj[field]), value)
    else:
      return match_exact(str(obj[field]), value)

  return True